import logging
import operator
import datetime

from koapy.grpc import KiwoomOpenApiService_pb2
from koapy.grpc.event.BaseKiwoomOpenApiEventHandler import BaseKiwoomOpenApiEventHandler
from koapy.openapi.KiwoomOpenApiError import KiwoomOpenApiError
from koapy.openapi.TrInfo import TrInfo
from koapy.openapi.RealType import RealType

from koapy.utils.notimplemented import isimplemented

class KiwoomOpenApiLoggingEventHandler(BaseKiwoomOpenApiEventHandler):

    def OnReceiveTrData(self, scrnno, rqname, trcode, recordname, prevnext, _datalength, _errorcode, _message, _splmmsg):
        logging.debug('OnReceiveTrData(%r, %r, %r, %r, %r)', scrnno, rqname, trcode, recordname, prevnext)

    def OnReceiveRealData(self, code, realtype, realdata):
        logging.debug('OnReceiveRealData(%r, %r, %r)', code, realtype, realdata)
        if code == '09' and realtype == '장시작시간':
            signal_type = self.control.GetCommRealData(code, 215)
            current_time = self.control.GetCommRealData(code, 20)
            estimated_remaining_time = self.control.GetCommRealData(code, 214)
            signal_type_msg = {
                # 아래는 문서에서 확인 가능
                '0': '장시작전(동시호가시작,이후1분단위)',
                '2': '장종료전(동시호가시작,이후1분단위)',
                '3': '장시작(동시호가종료)',
                '4': '장종료(동시호가종료)',
                '8': '장종료(시간내용없음)',
                '9': '장마감(시간내용없음)',
                # 이후는 추정
                's': '선물옵션장종료전(동시호가시작)', # 17번반복
                'a': '장후시간외종가시작',
                'e': '선물옵션장종료(동시호가종료)', # 17번반복
                'b': '장후시간외종가종료',
                'c': '시간외단일가시작',
                'd': '시간외단일가종료',
            }.get(signal_type, '알수없음')
            if signal_type not in ['8', '9']:
                current_time = datetime.datetime.strptime(current_time, '%H%M%S')
            else:
                current_time = datetime.datetime.now()
            ert = datetime.datetime.strptime(estimated_remaining_time, '%H%M%S')
            estimated_remaining_time = datetime.timedelta(hours=ert.hour, minutes=ert.minute, seconds=ert.second)
            logging.debug('%s, %s remaining', signal_type_msg, estimated_remaining_time)

    def OnReceiveMsg(self, scrnno, rqname, trcode, msg):
        """
        [OnReceiveMsg()이벤트]

          OnReceiveMsg(
          BSTR sScrNo,   // 화면번호
          BSTR sRQName,  // 사용자 구분명
          BSTR sTrCode,  // TR이름
          BSTR sMsg     // 서버에서 전달하는 메시지
          )

          서버통신 후 수신한 메시지를 알려줍니다.
          메시지에는 6자리 코드번호가 포함되는데 이 코드번호는 통보없이 수시로 변경될 수 있습니다. 따라서 주문이나 오류관련처리를
          이 코드번호로 분류하시면 안됩니다.
        """
        logging.debug('OnReceiveMsg(%r, %r, %r, %r)', scrnno, rqname, trcode, msg)

        if msg == '전문 처리 실패(-22)':
            logging.warning('Server might have ended connection abruptly')

    def OnReceiveChejanData(self, gubun, itemcnt, fidlist):
        logging.debug('OnReceiveChejanData(%r, %r, %r)', gubun, itemcnt, fidlist)

    def OnEventConnect(self, errcode):
        logging.debug('OnEventConnect(%r)', errcode)

    def OnReceiveRealCondition(self, code, condition_type, condition_name, condition_index):
        logging.debug('OnReceiveRealCondition(%r, %r, %r, %r)', code, condition_type, condition_name, condition_index)

    def OnReceiveTrCondition(self, scrnno, codelist, condition_name, index, prevnext):
        logging.debug('OnReceiveTrCondition(%r, %r, %r, %r, %r)', scrnno, codelist, condition_name, index, prevnext)

    def OnReceiveConditionVer(self, ret, msg):
        logging.debug('OnReceiveConditionVer(%r, %r)', ret, msg)

class KiwoomOpenApiAllEventHandler(BaseKiwoomOpenApiEventHandler):

    def OnReceiveTrData(self, scrnno, rqname, trcode, recordname, prevnext, _datalength, _errorcode, _message, _splmmsg):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveTrData' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = scrnno
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = rqname
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = trcode
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = recordname
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = prevnext
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveRealData(self, code, realtype, realdata):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveRealData' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = code
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = realtype
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = realdata
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveMsg(self, scrnno, rqname, trcode, msg):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveMsg' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = scrnno
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = rqname
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = trcode
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = msg
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveChejanData(self, gubun, itemcnt, fidlist):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveChejanData' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = gubun
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = itemcnt
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = fidlist
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnEventConnect(self, errcode):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnEventConnect' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = errcode
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveRealCondition(self, code, condition_type, condition_name, condition_index):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveRealCondition' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = code
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = condition_type
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = condition_name
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = condition_index
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveTrCondition(self, scrnno, codelist, condition_name, index, prevnext):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveTrCondition' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = scrnno
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = codelist
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = condition_name
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = index
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = prevnext
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

    def OnReceiveConditionVer(self, ret, msg):
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveConditionVer' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = ret
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = msg
        self.observer.on_next(response.listen_response) # pylint: disable=no-member

class KiwoomOpenApiSomeEventHandler(KiwoomOpenApiAllEventHandler):

    def __init__(self, control, request):
        super().__init__(control)
        self._request = request

    def slots(self):
        names = self.names()
        slots = [getattr(self, name) for name in names]
        names_and_slots_implemented = [(name, slot) for name, slot in zip(names, slots) if isimplemented(slot) and name in self._request.slots]
        return names_and_slots_implemented

class KiwoomOpenApiLoginEventHandler(BaseKiwoomOpenApiEventHandler):

    def __init__(self, control, request):
        super().__init__(control)
        self._request = request

    def OnEventConnect(self, errcode):
        if errcode < 0:
            error = KiwoomOpenApiError(errcode)
            self.observer.on_error(error)
        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnEventConnect' # pylint: disable=no-member
        response.listen_response.arguments.add().long_value = errcode # pylint: disable=no-member
        self.observer.on_next(response)
        self.observer.on_completed()

class KiwoomOpenApiTrEventHandler(BaseKiwoomOpenApiEventHandler):

    def __init__(self, control, request):
        super().__init__(control)
        self._request = request

        self._rqname = request.request_name
        self._trcode = request.transaction_code
        self._scrnno = request.screen_no
        self._inputs = request.inputs

        self._trinfo = TrInfo.get_trinfo_by_code(self._trcode)

        if self._trinfo is None:
            logging.error('Cannot find names for trcode %s', self._trinfo)

        self._input_code = self._inputs.get('종목코드')

        self._single_names = self._trinfo.get_single_output_names()
        self._multi_names = self._trinfo.get_multi_output_names()

        stop_condition = request.stop_condition
        stop_condition_is_valid = all([
            stop_condition is not None,
            stop_condition.name is not None,
            len(stop_condition.name) > 0,
            stop_condition.name in self._multi_names])

        if stop_condition_is_valid:
            column_index_to_check = self._multi_names.index(stop_condition.name)
            comparator = {
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.LESS_THAN_OR_EQUAL_TO: operator.le,
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.LESS_THAN: operator.lt,
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.GREATER_THAN_OR_EQUAL_TO: operator.ge,
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.GREATER_THAN: operator.gt,
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.EQUAL_TO: operator.eq,
                KiwoomOpenApiService_pb2.TransactionStopConditionCompartor.NOT_EQUAL_TO: operator.ne,
            }.get(stop_condition.comparator, operator.le)
            def is_stop_condition(row):
                return comparator(row[column_index_to_check], stop_condition.value)
        else:
            def is_stop_condition(_):
                return False

        self._is_stop_condition = is_stop_condition

    def OnReceiveTrData(self, scrnno, rqname, trcode, recordname, prevnext, datalength, errorcode, message, splmmsg):
        if (rqname, trcode, scrnno) == (self._rqname, self._trcode, self._scrnno):
            response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
            response.listen_response.name = 'OnReceiveTrData' # pylint: disable=no-member
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = scrnno
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = rqname
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = trcode
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = recordname
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = prevnext

            should_stop = prevnext in ['', '0']
            repeat_cnt = self.control.GetRepeatCnt(trcode, recordname)

            if len(self._single_names) > 0:
                values = [self.control.GetCommData(trcode, recordname, 0, name).strip() for name in self._single_names]
                response.listen_response.single_data.names.extend(self._single_names) # pylint: disable=no-member
                response.listen_response.single_data.values.extend(values) # pylint: disable=no-member

            if len(self._multi_names) > 0 and repeat_cnt > 0:
                rows = [[self.control.GetCommData(trcode, recordname, i, name).strip() for name in self._multi_names] for i in range(repeat_cnt)]
                response.listen_response.multi_data.names.extend(self._multi_names) # pylint: disable=no-member
                for row in rows:
                    if self._is_stop_condition(row):
                        should_stop = True
                        break
                    response.listen_response.multi_data.values.add().values.extend(row) # pylint: disable=no-member

            self.observer.on_next(response)

            if should_stop:
                self.observer.on_completed()
                return
            else:
                for k, v in self._inputs.items():
                    self.control.SetInputValue(k, v)
                try:
                    self.control.RateLimitedCommRqData(rqname, trcode, int(prevnext), scrnno, self._inputs)
                except KiwoomOpenApiError as e:
                    self.observer.on_error(e)
                    return

    def OnEventConnect(self, errcode):
        if errcode < 0:
            error = KiwoomOpenApiError(errcode)
            self.observer.on_error(error)
            return

class KiwoomOpenApiOrderEventHandler(BaseKiwoomOpenApiEventHandler):

    def __init__(self, control, request):
        super().__init__(control)
        self._request = request

        self._rqname = request.request_name
        self._scrnno = request.screen_no
        self._accno = request.account_no
        self._ordertype = request.order_type
        self._code = request.code
        self._qty = request.quantity
        self._price = request.price
        self._hogagb = request.quote_type
        self._orgorderno = request.original_order_no

        self._order_no = None
        self._should_stop = False

        self._single_names = ['주문번호']
        self._multi_names = []

        self._is_stop_condition = lambda row: False
        self._inputs = {}

    def OnReceiveMsg(self, scrnno, rqname, trcode, msg):
        if (rqname, scrnno) == (self._rqname, self._scrnno):
            response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
            response.listen_response.name = 'OnReceiveMsg' # pylint: disable=no-member
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = rqname
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = trcode
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = msg

            self.observer.on_next(response)

            # 아래는 개발과정에서 확인용도
            if trcode == 'KOA_NORMAL_BUY_KP_ORD':
                if msg == '[00Z218] 모의투자 장종료 상태입니다': # 메시지의 코드로 판단하는건 위험함 (게다가 모의투자만 해당)
                    logging.warning('Market is closed')
                elif msg == '[00Z217] 모의투자 장시작전입니다':
                    logging.warning('Market is before opening')
                elif msg == '[00Z237] 모의투자 단가를 입력하지 않는 호가입니다.':
                    logging.warning('Price is given but should not')
                elif msg == '[00Z112] 모의투자 정상처리 되었습니다':
                    logging.debug('Processed successfully')
            elif trcode == 'KOA_NORMAL_KP_CANCEL':
                if msg == '[00Z924] 모의투자 취소수량이 취소가능수량을 초과합니다':
                    logging.warning('Not enough amount to cancel')

    def OnReceiveTrData(self, scrnno, rqname, trcode, recordname, prevnext, datalength, errorcode, message, splmmsg):
        if (rqname, scrnno) == (self._rqname, self._scrnno):
            response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
            response.listen_response.name = 'OnReceiveTrData' # pylint: disable=no-member
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = scrnno
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = rqname
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = trcode
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = recordname
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = prevnext

            self._order_no = self.control.GetCommData(trcode, recordname, 0, '주문번호').strip()

            should_stop = prevnext in ['', '0']
            repeat_cnt = self.control.GetRepeatCnt(trcode, recordname)

            if len(self._single_names) > 0:
                values = [self.control.GetCommData(trcode, recordname, 0, name).strip() for name in self._single_names]
                response.listen_response.single_data.names.extend(self._single_names) # pylint: disable=no-member
                response.listen_response.single_data.values.extend(values) # pylint: disable=no-member

            if len(self._multi_names) > 0 and repeat_cnt > 0:
                rows = [[self.control.GetCommData(trcode, recordname, i, name).strip() for name in self._multi_names] for i in range(repeat_cnt)]
                response.listen_response.multi_data.names.extend(self._multi_names) # pylint: disable=no-member
                for row in rows:
                    if self._is_stop_condition(row):
                        should_stop = True
                        break
                    response.listen_response.multi_data.values.add().values.extend(row) # pylint: disable=no-member

            self.observer.on_next(response)

            if not self._order_no:
                self.observer.on_completed()
                return
            elif not should_stop:
                for k, v in self._inputs.items():
                    self.control.SetInputValue(k, v)
                try:
                    self.control.CommRqDataAndCheck(rqname, trcode, int(prevnext), scrnno, self._inputs)
                except KiwoomOpenApiError as e:
                    self.observer.on_error(e)
                    return

    def OnReceiveChejanData(self, gubun, itemcnt, fidlist):
        fids = fidlist.rstrip(';')
        fids = fids.split(';') if fids else []
        fids = [int(fid) for fid in fids]

        assert itemcnt == len(fids)

        names = [RealType.Fid.get_name_by_fid(fid, str(fid)) for fid in fids]
        values = [self.control.GetChejanData(fid).strip() for fid in fids]

        response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
        response.listen_response.name = 'OnReceiveChejanData' # pylint: disable=no-member
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = gubun
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.long_value = itemcnt
        argument = response.listen_response.arguments.add() # pylint: disable=no-member
        argument.string_value = fidlist

        response.listen_response.single_data.names.extend(names) # pylint: disable=no-member
        response.listen_response.single_data.values.extend(values) # pylint: disable=no-member

        if gubun == '0': # 접수와 체결시 (+ 취소 확인)
            accno = self.control.GetChejanData(9201).strip()
            scrnno = self.control.GetChejanData(920).strip()
            order_no = self.control.GetChejanData(9203).strip()
            if (scrnno, accno, order_no) == (self._scrnno, self._accno, self._order_no):
                self.observer.on_next(response)
                status = self.control.GetChejanData(913).strip()
                if status == '접수':
                    pass
                elif status == '체결':
                    orders_left = self.control.GetChejanData(902).strip()
                    orders_left = int(orders_left) if orders_left.isdigit() else 0
                    if orders_left == 0:
                        self._should_stop = True # 미체결수량이 더이상 없다면 이후 잔고 이벤트 후 종료
                elif status == '확인':
                    self.observer.on_completed()
                    return
        elif gubun == '1': # 국내주식 잔고전달
            accno = self.control.GetChejanData(9201).strip()
            code = self.control.GetChejanData(9001).strip()
            if accno == self._accno and code.endswith(self._code): # code 비교시에 앞에 prefix 가 붙어오기 때문에 endswith 으로 비교해야됨
                self.observer.on_next(response)
                if self._should_stop: # 미체결수량이 더이상 없다면 이후 잔고 이벤트 후 종료
                    self.observer.on_completed()
                    return
        elif gubun == '4': # 파생 잔고전달
            accno = self.control.GetChejanData(9201).strip()
            code = self.control.GetChejanData(9001).strip()
            if accno == self._accno and code.endswith(self._code):
                self.observer.on_next(response)
                if self._should_stop: # 미체결수량이 더이상 없다면 이후 잔고 이벤트 후 종료
                    self.observer.on_completed()
                    return
        else:
            logging.error('Unexpected gubun value: %r', gubun)
            e = ValueError('Unexpected gubun value: %r' % gubun)
            self.observer.on_error(e)
            return

    def OnEventConnect(self, errcode):
        if errcode < 0:
            error = KiwoomOpenApiError(errcode)
            self.observer.on_error(error)
            return

class KiwoomOpenApiRealEventHandler(BaseKiwoomOpenApiEventHandler):

    def __init__(self, control, request):
        super().__init__(control)
        self._request = request

        self._screen_no = request.screen_no
        self._code_list = request.code_list
        self._fid_list = request.fid_list
        self._real_type = request.real_type

        self._infer_fids = request.flags.infer_fids
        self._readable_names = request.flags.readable_names
        self._fast_parse = request.flags.fast_parse

    def OnReceiveRealData(self, code, realtype, realdata):
        if code in self._code_list:
            response = KiwoomOpenApiService_pb2.CustomCallAndListenResponse()
            response.listen_response.name = 'OnReceiveRealData' # pylint: disable=no-member
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = code
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = realtype
            argument = response.listen_response.arguments.add() # pylint: disable=no-member
            argument.string_value = realdata

            if self._infer_fids:
                fids = RealType.get_fids_by_realtype(realtype)
            else:
                fids = self._fid_list

            if self._readable_names:
                names = [RealType.Fid.get_name_by_fid(fid, str(fid)) for fid in fids]
            else:
                names = [str(fid) for fid in fids]

            if self._infer_fids and self._fast_parse:
                values = realdata.split('\t')
            else:
                values = [self.control.GetCommRealData(code, fid) for fid in fids]

            assert len(names) == len(values)

            response.listen_response.single_data.names.extend(names) # pylint: disable=no-member
            response.listen_response.single_data.values.extend(values) # pylint: disable=no-member

            self.observer.on_next(response)

    def OnEventConnect(self, errcode):
        if errcode < 0:
            error = KiwoomOpenApiError(errcode)
            self.observer.on_error(error)
            return



import io
import json

class JsonSerializable:

    def to_dict(self):
        return dict(self.__dict__)

    @classmethod
    def from_dict(cls, dic):
        output = cls()
        for name in output.__dict__:
            setattr(output, name, dic.get(name))
        return output

    def to_json(self, f=None):
        if f is None:
            return json.dumps(self.to_dict())
        elif isinstance(f, str):
            with open(f, 'w') as f:
                return json.dump(self.to_dict(), f)
        elif isinstance(f, io.TextIOBase):
            return json.dump(self.to_dict(), f)
        else:
            raise ValueError('Unsupported argument type: %s' % type(f))

    @classmethod
    def from_json(cls, jsn):
        if isinstance(jsn, str):
            dic = json.loads(jsn)
        elif isinstance(jsn, io.TextIOBase):
            dic = json.load(jsn)
        else:
            raise ValueError('Unsupported argument type: %s' % type(jsn))
        return cls.from_dict(dic)



import os
import logging.config

from pyhocon import ConfigFactory

def read_config(filename):
    return ConfigFactory.parse_file(filename)

config_filename_cadidates = [
    'koapy.conf',
    '.koapy.conf',
]

default_config_filename = 'config.conf'
file_directory = os.path.dirname(os.path.realpath(__file__))
default_config_filepath = os.path.join(file_directory, default_config_filename)
default_config = read_config(default_config_filepath)

empty_config = ConfigFactory.from_dict({})
current_directory_config = empty_config

cwd = os.getcwd()
for config_filename in config_filename_cadidates:
    config_path = os.path.join(cwd, config_filename)
    if os.path.exists(config_path):
        current_directory_config = read_config(config_path)
        break

config = current_directory_config.with_fallback(default_config)

logging.config.dictConfig(config.get('koapy.utils.logging'))



import io
import tqdm

class TqdmStream(io.RawIOBase):

    def __init__(self, file=None):
        super().__init__()
        self._file = file

    def __getattr__(self, name):
        return getattr(self._file, name)

    def fileno(self):
        if self._file is None:
            raise OSError()
        return self._file.fileno()

    def seek(self, offset, whence=io.SEEK_SET):
        if self._file is None:
            raise OSError()
        return self._file.seek(offset, whence)

    def truncate(self):
        if self._file is None:
            raise OSError()
        return self._file.truncate()

    def write(self, b):
        if len(b.rstrip()) > 0:
            tqdm.tqdm.write(b, file=self._file, end='')



import logging

class KiwoomOpenApiDynamicCallable:

    def __init__(self, control, name):
        self._control = control
        self._name = name

    @classmethod
    def _getInputParameterType(cls, param):
        if isinstance(param, str):
            return 'const QString&'
        elif isinstance(param, int):
            return 'int' # giving 'long' breaks behavior of CommRqData()
        else:
            raise TypeError('Cannot recognize type for input param: %s' % type(param))

    @classmethod
    def _createDynamicCallFunction(cls, name, args):
        return '%s(%s)' % (name, ', '.join(map(cls._getInputParameterType, args)))

    def _staticCall(self, *args, **kwargs):
        return getattr(self._control, self._name)(*args, **kwargs)

    def _dynamicCall(self, *args, **kwargs):
        if kwargs:
            logging.warning('Ignoring given kwargs..')
        function = self._createDynamicCallFunction(self._name, args)
        # logging.debug('Calling dynamicCall(%r, %r)', function, args)
        if len(args) > 0:
            # giving args as tuple raises an error, ex) _.dynamicCall('...', args)
            # giving unpacked args has length limitation, ex) _.dynamicCall('...', *args)
            return self._control.dynamicCall(function, list(args))
        else:
            return self._control.dynamicCall(function)

    def __call__(self, *args, **kwargs):
        return self._dynamicCall(*args, **kwargs)



import logging

from PyQt5.QtWidgets import QWidget
from PyQt5.QAxContainer import QAxWidget
from PyQt5.QtCore import QEvent, Qt

from koapy.pyqt5.KiwoomOpenApiDynamicCallable import KiwoomOpenApiDynamicCallable
from koapy.pyqt5.KiwoomOpenApiSignalConnector import KiwoomOpenApiSignalConnector
from koapy.utils.rate_limiting.RateLimiter import SimpleRateLimiter
from koapy.openapi.KiwoomOpenApiError import KiwoomOpenApiError
from koapy.grpc.event.KiwoomOpenApiEventHandler import KiwoomOpenApiLoggingEventHandler
from koapy.grpc.event.KiwoomOpenApiEventHandlerFunctions import KiwoomOpenApiEventHandlerFunctions

class KiwoomOpenApiQAxWidget(QWidget):

    CLSID = '{A1574A0D-6BFA-4BD7-9020-DED88711818D}'
    PROGID = 'KHOPENAPI.KHOpenApiCtrl.1'

    try:
        import pythoncom
        from pywintypes import com_error as ComError
    except ImportError:
        pass
    else:
        try:
            PROGID = pythoncom.ProgIDFromCLSID(CLSID) or PROGID
        except ComError:
            pass
        finally:
            del pythoncom
            del ComError

    CONTROL_NAME_KWARG_KEY = 'c'

    def __init__(self, *args, **kwargs):
        super_args = args
        super_kwargs = kwargs

        control_name = self.PROGID

        if len(args) > 0 and isinstance(args[0], str):
            super_args = args[1:]
            control_name = args[0]
        elif self.CONTROL_NAME_KWARG_KEY in kwargs:
            super_kwargs = {k:v for k, v in kwargs if k != self.CONTROL_NAME_KWARG_KEY}
            control_name = kwargs[self.CONTROL_NAME_KWARG_KEY]

        super().__init__(*super_args, **super_kwargs)

        self._ax = QAxWidget(control_name, self)
        self._signals = {}
        self._event_logger = KiwoomOpenApiLoggingEventHandler(self)

        for event_name in [name for name in dir(KiwoomOpenApiEventHandlerFunctions) if name.startswith('On')]:
            # 아래처럼 중간의 프록시 객체 (KiwoomOpenApiSignalConnector) 를 넣지
            # 않으면 외부에서 동적으로 connect 를 할 수가 없어보임 (하더라도
            # 제대로 이벤트를 받지 못함)
            connector = KiwoomOpenApiSignalConnector()
            connector.connect(getattr(self._event_logger, event_name))
            getattr(self._ax, event_name).connect(connector)
            self._signals[event_name] = connector
            # 아래는 기존에 시도해서 실패한 구현 (제대로 이벤트를 받지 못함)
            """
            connector = getattr(self._ax, event_name)
            connector.connect(getattr(self._event_logger, event_name))
            self._signals[event_name] = connector
            """

    def __getattr__(self, name):
        result = getattr(self._ax, name)

        if type(result).__name__ == 'pyqtMethodProxy':
            return KiwoomOpenApiDynamicCallable(self._ax, name)
        elif name.startswith('On') and name in self._signals:
            return self._signals[name]
        else:
            return result

    def changeEvent(self, event):
        if event.type() == QEvent.WindowStateChange:
            if self.windowState() & Qt.WindowMinimized:
                event.accept()

    def closeEvent(self, event):
        self.hide()
        event.ignore()

    @SimpleRateLimiter(period=4, calls=1) # 그냥 1초당 5회로하면 장기적으로 결국 막히기 때문에 4초당 1회로 제한 (3초당 1회부턴 제한걸림)
    def RateLimitedCommRqData(self, rqname, trcode, prevnext, scrnno, inputs=None):
        """
        [OpenAPI 게시판]
          https://bbn.kiwoom.com/bbn.openAPIQnaBbsList.do

        [조회횟수 제한 관련 가이드]
          - 1초당 5회 조회를 1번 발생시킨 경우 : 17초대기
          - 1초당 5회 조회를 5연속 발생시킨 경우 : 90초대기
          - 1초당 5회 조회를 10연속 발생시킨 경우 : 3분(180초)대기
        """
        prevnext = int(prevnext) # ensure prevnext is int
        code = self.CommRqData(rqname, trcode, prevnext, scrnno)
        spec = 'CommRqData(%r, %r, %r, %r)' % (rqname, trcode, prevnext, scrnno)

        if inputs is not None:
            spec += ' with inputs %r' % inputs

        if code == KiwoomOpenApiError.OP_ERR_NONE:
            message = 'CommRqData() was successful; ' + spec
            logging.debug(message)
        elif code == KiwoomOpenApiError.OP_ERR_SISE_OVERFLOW:
            message = 'CommRqData() was rejected due to massive request; ' + spec
            logging.error(message)
            raise KiwoomOpenApiError(code)
        elif code == KiwoomOpenApiError.OP_ERR_ORD_WRONG_INPUT:
            message = 'CommRqData() failed due to wrong input, check if input was correctly set; ' + spec
            logging.error(message)
            raise KiwoomOpenApiError(code)
        elif code in (KiwoomOpenApiError.OP_ERR_RQ_STRUCT_FAIL, KiwoomOpenApiError.OP_ERR_RQ_STRING_FAIL):
            message = 'CommRqData() request was invalid; ' + spec
            logging.error(message)
            raise KiwoomOpenApiError(code)
        else:
            message = 'Unknown error occured during CommRqData() request; ' + spec
            korean_message = KiwoomOpenApiError.get_error_message_by_code(code)
            if korean_message is not None:
                message += '; Korean error message: ' +  korean_message
            logging.error(message)
            raise KiwoomOpenApiError(code)

        return code



import itertools

def chunk(iterable, n):
    fillvalue = object()
    args = [iter(iterable)] * n
    it = itertools.zip_longest(fillvalue=fillvalue, *args)
    it = map(lambda items: [item for item in items if item is not fillvalue], it)
    return it



# Generated by the gRPC Python protocol compiler plugin. DO NOT EDIT!
"""Client and server classes corresponding to protobuf-defined services."""
import grpc

from koapy.grpc import KiwoomOpenApiService_pb2 as koapy_dot_grpc_dot_KiwoomOpenApiService__pb2


class KiwoomOpenApiServiceStub(object):
    """Missing associated documentation comment in .proto file."""

    def __init__(self, channel):
        """Constructor.

        Args:
            channel: A grpc.Channel.
        """
        self.Call = channel.unary_unary(
                '/koapy.grpc.KiwoomOpenApiService/Call',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallResponse.FromString,
                )
        self.Listen = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/Listen',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.ListenRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.FromString,
                )
        self.StopListen = channel.unary_unary(
                '/koapy.grpc.KiwoomOpenApiService/StopListen',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenResponse.FromString,
                )
        self.CustomListen = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/CustomListen',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.FromString,
                )
        self.CustomCallAndListen = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/CustomCallAndListen',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
                )
        self.LoginCall = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/LoginCall',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.LoginRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
                )
        self.TransactionCall = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/TransactionCall',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.TransactionRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
                )
        self.OrderCall = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/OrderCall',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.OrderRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
                )
        self.RealCall = channel.unary_stream(
                '/koapy.grpc.KiwoomOpenApiService/RealCall',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.RealRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
                )
        self.SetLogLevel = channel.unary_unary(
                '/koapy.grpc.KiwoomOpenApiService/SetLogLevel',
                request_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelRequest.SerializeToString,
                response_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelResponse.FromString,
                )


class KiwoomOpenApiServiceServicer(object):
    """Missing associated documentation comment in .proto file."""

    def Call(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def Listen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def StopListen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CustomListen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def CustomCallAndListen(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def LoginCall(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def TransactionCall(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def OrderCall(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def RealCall(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')

    def SetLogLevel(self, request, context):
        """Missing associated documentation comment in .proto file."""
        context.set_code(grpc.StatusCode.UNIMPLEMENTED)
        context.set_details('Method not implemented!')
        raise NotImplementedError('Method not implemented!')


def add_KiwoomOpenApiServiceServicer_to_server(servicer, server):
    rpc_method_handlers = {
            'Call': grpc.unary_unary_rpc_method_handler(
                    servicer.Call,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallResponse.SerializeToString,
            ),
            'Listen': grpc.unary_stream_rpc_method_handler(
                    servicer.Listen,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.ListenRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.SerializeToString,
            ),
            'StopListen': grpc.unary_unary_rpc_method_handler(
                    servicer.StopListen,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenResponse.SerializeToString,
            ),
            'CustomListen': grpc.unary_stream_rpc_method_handler(
                    servicer.CustomListen,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.SerializeToString,
            ),
            'CustomCallAndListen': grpc.unary_stream_rpc_method_handler(
                    servicer.CustomCallAndListen,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.SerializeToString,
            ),
            'LoginCall': grpc.unary_stream_rpc_method_handler(
                    servicer.LoginCall,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.LoginRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.SerializeToString,
            ),
            'TransactionCall': grpc.unary_stream_rpc_method_handler(
                    servicer.TransactionCall,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.TransactionRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.SerializeToString,
            ),
            'OrderCall': grpc.unary_stream_rpc_method_handler(
                    servicer.OrderCall,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.OrderRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.SerializeToString,
            ),
            'RealCall': grpc.unary_stream_rpc_method_handler(
                    servicer.RealCall,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.RealRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.SerializeToString,
            ),
            'SetLogLevel': grpc.unary_unary_rpc_method_handler(
                    servicer.SetLogLevel,
                    request_deserializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelRequest.FromString,
                    response_serializer=koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelResponse.SerializeToString,
            ),
    }
    generic_handler = grpc.method_handlers_generic_handler(
            'koapy.grpc.KiwoomOpenApiService', rpc_method_handlers)
    server.add_generic_rpc_handlers((generic_handler,))


 # This class is part of an EXPERIMENTAL API.
class KiwoomOpenApiService(object):
    """Missing associated documentation comment in .proto file."""

    @staticmethod
    def Call(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/koapy.grpc.KiwoomOpenApiService/Call',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CallResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def Listen(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/Listen',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.ListenRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def StopListen(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/koapy.grpc.KiwoomOpenApiService/StopListen',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.StopListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CustomListen(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/CustomListen',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def CustomCallAndListen(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/CustomCallAndListen',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def LoginCall(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/LoginCall',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.LoginRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def TransactionCall(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/TransactionCall',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.TransactionRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def OrderCall(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/OrderCall',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.OrderRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def RealCall(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_stream(request, target, '/koapy.grpc.KiwoomOpenApiService/RealCall',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.RealRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.CustomCallAndListenResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)

    @staticmethod
    def SetLogLevel(request,
            target,
            options=(),
            channel_credentials=None,
            call_credentials=None,
            insecure=False,
            compression=None,
            wait_for_ready=None,
            timeout=None,
            metadata=None):
        return grpc.experimental.unary_unary(request, target, '/koapy.grpc.KiwoomOpenApiService/SetLogLevel',
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelRequest.SerializeToString,
            koapy_dot_grpc_dot_KiwoomOpenApiService__pb2.SetLogLevelResponse.FromString,
            options, channel_credentials,
            insecure, call_credentials, compression, wait_for_ready, timeout, metadata)



from abc import ABC
from abc import abstractmethod

class Observer(ABC):

    @abstractmethod
    def on_next(self, value):
        raise NotImplementedError

    @abstractmethod
    def on_error(self, error):
        raise NotImplementedError

    @abstractmethod
    def on_completed(self):
        raise NotImplementedError



def notimplemented(func):
    func.__isnotimplemented__ = True
    return func

def isnotimplemented(func):
    return getattr(func, '__isnotimplented__', False)

def isimplemented(func):
    return not isnotimplemented(func)
    


"""Console script for koapy."""

import os
import locale
import logging

import click

import koapy
from koapy.utils.logging import set_verbosity

CONTEXT_SETTINGS = dict(help_option_names=['-h', '--help'])

client_check_timeout = 3

def fail_with_usage(message=None):
    ctx = click.get_current_context()
    if message is not None:
        click.UsageError(message).show()
        click.echo()
    click.echo(ctx.get_help())
    ctx.exit(1)

@click.group(context_settings=CONTEXT_SETTINGS)
@click.version_option(koapy.__version__, '-V', '--version')
def cli():
    pass

@cli.command(context_settings=CONTEXT_SETTINGS, short_help='Start grpc server with tray application.')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
@click.option('-v', '--verbose', count=True, default=3, help='Verbosity.')
@click.argument('args', nargs=-1)
def serve(port, verbose, args):
    """
    ARGS are passed to QApplication.
    """
    args = ['--port', port]
    if verbose > 0:
        args.append('-' + 'v' * verbose)
    args += list(args)
    from koapy.pyqt5.KiwoomOpenApiTrayApplication import KiwoomOpenApiTrayApplication
    KiwoomOpenApiTrayApplication.main(args)

@cli.group(short_help='Get various types of data.')
def get():
    pass

market_codes = [
    '0',
    '10',
    '3',
    '8',
    '50',
    '4',
    '5',
    '6',
    '9',
    '30',
    'all',
]

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get stock codes.')
@click.option('-m', '--market', 'markets', metavar='MARKET', multiple=True, type=click.Choice(market_codes, case_sensitive=False), help='Stock market code to get. Can set multiple times.')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
def code(markets, port):
    """
    \b
    Possible market codes are:
      0 : 장내
      10 : 코스닥
      3 : ELW
      8 : ETF
      50 : KONEX
      4 : 뮤추얼펀드
      5 : 신주인수권
      6 : 리츠
      9 : 하이얼펀드
      30 : K-OTC

    \b
    Possible market code aliases are:
      all: All possible market codes.
    """

    if (markets) == (tuple()):
        fail_with_usage()

    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext

    if 'all' in markets:
        markets = market_codes

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout) as context:
        context.EnsureConnected()

        codes = set()

        for market in markets:
            codes = codes.union(set(context.GetCodeListByMarketAsList(market)))

        codes = sorted(list(codes))

        for code in codes:
            click.echo(code)

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get name for stock codes.')
@click.option('-c', '--code', 'codes', metavar='CODE', multiple=True, help='Stock code to get. Can set multiple times.')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
def name(codes, port):
    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout) as context:
        context.EnsureConnected()

        def get_codes():
            if codes:
                if '-' in codes:
                    with click.open_file('-', 'r') as f:
                        for code in f:
                            yield code.strip()
                else:
                    for code in codes:
                        yield code
            else:
                while True:
                    try:
                        code = click.prompt('code', prompt_suffix=' >>> ')
                        code = code.strip()
                        if code == 'exit':
                            break
                        if code:
                            yield code
                    except EOFError:
                        break

        for code in get_codes():
            name = context.GetMasterCodeName(code)
            click.echo(name)

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get basic information of stocks.')
@click.option('-c', '--code', 'codes', metavar='CODE', multiple=True, help='Stock code to get. Can set multiple times.')
@click.option('-m', '--market', 'markets', metavar='MARKET', multiple=True, type=click.Choice(market_codes, case_sensitive=False), help='Stock market code to get. Alternative to --code. Can set multiple times.')
@click.option('-i', '--input', metavar='FILENAME', type=click.Path(), help='Text or excel file containing codes. Alternative to --code or --market.')
@click.option('-o', '--output', metavar='FILENAME', type=click.Path(), help="Output filename. Optional for single code (prints to console).")
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
@click.option('-v', '--verbose', count=True, help='Verbosity.')
def info(codes, markets, input, output, port, verbose):
    """
    \b
    Possible market codes are:
      0 : 장내
      10 : 코스닥
      3 : ELW
      8 : ETF
      50 : KONEX
      4 : 뮤추얼펀드
      5 : 신주인수권
      6 : 리츠
      9 : 하이얼펀드
      30 : K-OTC

    \b
    Possible market code aliases are:
      all: All possible market codes.
    """

    if (codes, markets, input, output) == (tuple(), tuple(), None, None):
        fail_with_usage()

    set_verbosity(verbose)

    codes_from_input = False
    codes_len = len(codes)

    if codes_len == 0 and len(markets) == 0:
        if input is None:
            fail_with_usage('Cannot specify codes.')
        if not os.path.exists(input):
            fail_with_usage('Given input does not exist.')

        codes_from_input = True

        if os.path.isfile(input):
            if input.endswith('.xlsx'):
                import pandas as pd
                df = pd.read_excel(input, dtype='object')
                code_column = '종목코드'
                if code_column in df:
                    codes = df[code_column]
                else:
                    codes = df.iloc[0]
                codes_len = len(codes)
            elif input.endswith('.txt'):
                with open(input) as f:
                    codes = [line.strip() for line in f]
                codes_len = len(codes)
            else:
                fail_with_usage('Unrecognized input type.')
        else:
            fail_with_usage('Unrecognized input type.')

    if output is None:
        if codes_len > 1 or codes_from_input:
            fail_with_usage('Output path is not specified.')
    else:
        if not output.endswith('.xlsx'):
            output += '.xlsx'

    import pandas as pd

    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout, verbosity=verbose) as context:
        context.EnsureConnected()

        if not codes_from_input and codes_len == 1:
            df = context.GetStockInfoAsDataFrame(codes)
            if not output:
                click.echo(df.iloc[0].to_markdown())
            else:
                df.to_excel(output, index=False)
        elif codes_len > 0:
            df = context.GetStockInfoAsDataFrame(codes)
            df.to_excel(output, index=False)
        elif len(markets) > 0:
            if 'all' in markets:
                markets = market_codes
            with pd.ExcelWriter(output) as writer: # pylint: disable=abstract-class-instantiated
                for market in markets:
                    codes = context.GetCodeListByMarketAsList(market)
                    df = context.GetStockInfoAsDataFrame(codes)
                    df.to_excel(writer, index=False, sheet_name=market)
        else:
            fail_with_usage('Cannot specify codes.')

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get daily OHLCV of stocks.')
@click.option('-c', '--code', 'codes', metavar='CODE', multiple=True, help='Stock code to get. Can set multiple times.')
@click.option('-i', '--input', metavar='FILENAME', type=click.Path(), help='Text or excel file containing codes. Alternative to --codes option.')
@click.option('-o', '--output', metavar='FOLDER|FILENAME', type=click.Path(), help='Output foldername or filename for single code. Files inside the folder would be named as CODE.xlsx. Defaults to current directory.')
@click.option('-s', '--start-date', metavar='YYYY-MM-DD', type=click.DateTime(formats=['%Y-%m-%d', '%Y%m%d']), help='Most recent date to get. Defaults to today or yesterday if market is open.')
@click.option('-e', '--end-date', metavar='YYYY-MM-DD', type=click.DateTime(formats=['%Y-%m-%d', '%Y%m%d']), help='Stops if reached, not included (optional).')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
@click.option('-v', '--verbose', count=True, help='Verbosity.')
def daily(codes, input, output, start_date, end_date, port, verbose):
    if (codes, input, output, start_date, end_date) == (tuple(), None, None, None, None):
        fail_with_usage()

    set_verbosity(verbose)

    codes_len = len(codes)

    if codes_len == 0:
        if input is None:
            fail_with_usage('Either code or input should be given.')
        if not os.path.exists(input):
            fail_with_usage('Given input does not exist.')

        if os.path.isfile(input):
            if input.endswith('.xlsx'):
                import pandas as pd
                df = pd.read_excel(input, dtype='object')
                code_column = '종목코드'
                if code_column in df:
                    codes = df[code_column]
                else:
                    codes = df.iloc[0]
                codes_len = len(codes)
            elif input.endswith('.txt'):
                with open(input) as f:
                    codes = [line.strip() for line in f]
                codes_len = len(codes)
            else:
                fail_with_usage('Unrecognized input type.')
        elif os.path.isdir(input):
            import re
            codes = [os.path.splitext(name)[0] for name in os.listdir(input) if name.endswith('.xlsx')]
            codes = [code for code in codes if re.match(r'[0-9A-Z]+', code)]
            codes_len = len(codes)

    if output is None:
        output = '.'

    if os.path.exists(output):
        if os.path.isdir(output):
            output_is_folder = True
        else:
            output_is_folder = False
    else:
        if output.endswith('/') or output.endswith(os.path.sep) or codes_len > 1:
            output_is_folder = True
        else:
            output_is_folder = False

    if output_is_folder:
        def filepath_for_code(code):
            if not os.path.exists(output):
                os.mkdir(output)
            return os.path.relpath(os.path.join(output, code + '.xlsx'))
    else:
        if not output.endswith('.xlsx'):
            output += '.xlsx'
        def filepath_for_code(_):
            return output

    import datetime
    import pandas as pd

    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout, verbosity=verbose) as context:
        context.EnsureConnected()

        for i, code in enumerate(codes):
            logging.info('Starting to get stock data for code: %s (%d/%d)', code, i+1, codes_len)
            filepath = filepath_for_code(code)
            if os.path.exists(filepath):
                df = pd.read_excel(filepath, dtype='object')
                last_date = df.loc[0, '일자']
                last_date = datetime.datetime.strptime(last_date, '%Y%m%d')
                logging.info('Found existing file %s, prepending from %s until %s', os.path.basename(filepath), start_date, last_date)
                df = pd.concat([context.GetDailyStockDataAsDataFrame(code, start_date, last_date), df], sort=False)
            else:
                df = context.GetDailyStockDataAsDataFrame(code, start_date, end_date)
            df.to_excel(filepath, index=False)
            logging.info('Saved stock data for code %s to %s', code, filepath)

minute_intervals = [
    '1',
    '3',
    '5',
    '10',
    '15',
    '30',
    '45',
    '60',
]

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get minute OHLCV of stocks.')
@click.option('-c', '--code', 'codes', metavar='CODE', multiple=True, help='Stock code to get. Can set multiple times.')
@click.option('-t', '--interval', metavar='INTERVAL', type=click.Choice(minute_intervals, case_sensitive=False), help='Minute interval. Possible values are [%s]' % '|'.join(minute_intervals))
@click.option('-i', '--input', metavar='FILENAME', type=click.Path(), help='Text or excel file containing codes. Alternative to --codes option.')
@click.option('-o', '--output', metavar='FOLDER|FILENAME', type=click.Path(), help='Output foldername or filename for single code. Files inside the folder would be named as CODE.xlsx. Defaults to current directory.')
@click.option('-s', '--start-date', metavar="YYYY-MM-DD['T'hh:mm:ss]", type=click.DateTime(formats=['%Y-%m-%d', '%Y%m%d', '%Y-%m-%dT%H:%M:%S', '%Y%m%d%H%M%S']), help='Most recent date to get. Defaults to today or yesterday if market is open.')
@click.option('-e', '--end-date', metavar="YYYY-MM-DD['T'hh:mm:ss]", type=click.DateTime(formats=['%Y-%m-%d', '%Y%m%d', '%Y-%m-%dT%H:%M:%S', '%Y%m%d%H%M%S']), help='Stops if reached, not included (optional).')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
@click.option('-v', '--verbose', count=True, help='Verbosity.')
def minute(codes, interval, input, output, start_date, end_date, port, verbose):
    if (codes, interval, input, output, start_date, end_date) == (tuple(), None, None, None, None, None):
        fail_with_usage()

    set_verbosity(verbose)

    if interval is None:
        fail_with_usage('Interval is not set.')

    codes_len = len(codes)

    if codes_len == 0:
        if input is None:
            fail_with_usage('Either code or input should be given.')
        if not os.path.exists(input):
            fail_with_usage('Given input does not exist.')

        if os.path.isfile(input):
            if input.endswith('.xlsx'):
                import pandas as pd
                df = pd.read_excel(input, dtype='object')
                code_column = '종목코드'
                if code_column in df:
                    codes = df[code_column]
                else:
                    codes = df.iloc[0]
                codes_len = len(codes)
            elif input.endswith('.txt'):
                with open(input) as f:
                    codes = [line.strip() for line in f]
                codes_len = len(codes)
            else:
                fail_with_usage('Unrecognized input type.')
        elif os.path.isdir(input):
            import re
            codes = [os.path.splitext(name)[0] for name in os.listdir(input) if name.endswith('.xlsx')]
            codes = [code for code in codes if re.match(r'[0-9A-Z]+', code)]
            codes_len = len(codes)

    if output is None:
        output = '.'

    if os.path.exists(output):
        if os.path.isdir(output):
            output_is_folder = True
        else:
            output_is_folder = False
    else:
        if output.endswith('/') or output.endswith(os.path.sep) or codes_len > 1:
            output_is_folder = True
        else:
            output_is_folder = False

    if output_is_folder:
        def filepath_for_code(code):
            if not os.path.exists(output):
                os.mkdir(output)
            return os.path.relpath(os.path.join(output, code + '.xlsx'))
    else:
        if not output.endswith('.xlsx'):
            output += '.xlsx'
        def filepath_for_code(_):
            return output

    import datetime
    import pandas as pd

    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout, verbosity=verbose) as context:
        context.EnsureConnected()

        for i, code in enumerate(codes):
            logging.info('Starting to get stock data for code: %s (%d/%d)', code, i+1, codes_len)
            filepath = filepath_for_code(code)
            if os.path.exists(filepath):
                df = pd.read_excel(filepath, dtype='object')
                last_date = df.loc[0, '체결시간']
                last_date = datetime.datetime.strptime(last_date, '%Y%m%d%H%M%S')
                logging.info('Found existing file %s, prepending from %s until %s', os.path.basename(filepath), start_date, last_date)
                df = pd.concat([context.GetMinuteStockDataAsDataFrame(code, interval, start_date, last_date), df], sort=False)
            else:
                df = context.GetMinuteStockDataAsDataFrame(code, interval, start_date, end_date)
            df.to_excel(filepath, index=False)
            logging.info('Saved stock data for code %s to %s', code, filepath)

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get TR info.')
@click.option('-t', '--trcode', 'trcodes', metavar='TRCODE', multiple=True, help='TR code to get (like opt100001).')
def trinfo(trcodes):
    from koapy.openapi.TrInfo import TrInfo

    def get_codes():
        if trcodes:
            if '-' in trcodes:
                with click.open_file('-', 'r') as f:
                    for code in f:
                        yield code.strip()
            else:
                for code in trcodes:
                    yield code
        else:
            while True:
                try:
                    code = click.prompt('trcode', prompt_suffix=' >>> ')
                    code = code.strip()
                    if code == 'exit':
                        break
                    if code:
                        yield code
                except EOFError:
                    break

    for trcode in get_codes():
        trinfo = TrInfo.get_trinfo_by_code(trcode)
        click.echo('[%s] : [%s]' % (trinfo.tr_code.upper(), trinfo.name))
        click.echo('  [INPUT]')
        for input in trinfo.inputs:
            click.echo('    %s' % input.name)
        if trinfo.single_outputs:
            click.echo('  [OUTPUT] [SINGLE DATA] : [%s]' % trinfo.single_outputs_name)
            for output in trinfo.single_outputs:
                click.echo('    %s' % output.name)
        if trinfo.multi_outputs:
            click.echo('  [OUTPUT] [MULTI DATA]  : [%s]' % trinfo.multi_outputs_name)
            for output in trinfo.multi_outputs:
                click.echo('    %s' % output.name)

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get real type info.')
@click.option('-t', '--realtype', 'realtypes', metavar='REALTYPE', multiple=True, help='Real type name to get (like 주식시세).')
def realtype(realtypes):
    from koapy.openapi.RealType import RealType

    def get_realtypes():
        if realtypes:
            if '-' in realtypes:
                with click.open_file('-', 'r') as f:
                    for realtype in f:
                        yield code.strip()
            else:
                for realtype in realtypes:
                    yield realtype
        else:
            while True:
                try:
                    realtype = click.prompt('realtype', prompt_suffix=' >>> ')
                    realtype = realtype.strip()
                    if realtype == 'exit':
                        break
                    if realtype:
                        yield realtype
                except EOFError:
                    break

    for realtype in get_realtypes():
        fids = RealType.get_fids_by_realtype(realtype)
        if fids:
            names = [RealType.Fid.get_name_by_fid(fid, str(fid)) for fid in fids]
            for fid, name in zip(fids, names):
                click.echo('  [%s] = %s' % (fid, name))

@get.command(context_settings=CONTEXT_SETTINGS, short_help='Get market closing dates.')
@click.option('-o', '--output', metavar='FILENAME', type=click.Path(), help='Output filename (optional).')
@click.option('-v', '--verbose', count=True, help='Verbosity.')
def closing(output, verbose):
    set_verbosity(verbose)
    if output is None:
        import pandas as pd
        from koapy.utils.krx.closing import get_krx_closing_dates_as_dict
        response = get_krx_closing_dates_as_dict()
        lang = locale.getdefaultlocale()[0]
        if lang == 'ko_KR':
            day_key = 'kr_dy_tp'
            columns = ['일자 및 요일', '요일구분', '비고']
        else:
            day_key = 'dy_tp_cd'
            columns = ['date', 'day of week', 'comment']
        data = []
        for closing_date in response['block1']:
            date = closing_date['calnd_dd_dy']
            day = closing_date[day_key].strip()
            name = closing_date['holdy_nm']
            data.append([date, day, name])
        df = pd.DataFrame.from_records(data, columns=columns)
        click.echo(df.to_markdown())
    else:
        from koapy.utils.krx.closing import download_krx_closing_dates_as_excel
        if output and not output.endswith('.xls'):
            output += '.xls'
        download_krx_closing_dates_as_excel(output)

@cli.command(context_settings=CONTEXT_SETTINGS, short_help='Watch realtime data.')
@click.option('-c', '--code', 'codes', metavar='CODE', multiple=True, help='Stock code to get. Can set multiple times.')
@click.option('-i', '--input', metavar='FILENAME', type=click.Path(), help='Text or excel file containing codes. Alternative to --codes option.')
@click.option('-f', '--fid', 'fids', metavar='FID', multiple=True, help='FID to get. Can set multiple times.')
@click.option('-t', '--realtype', metavar='REALTYPE', help='Real type name. Alternative to --fid.')
@click.option('-o', '--output', metavar='FILENAME', type=click.File('w', lazy=True), default='-', help='Output filename (optional).')
@click.option('-p', '--port', metavar='PORT', help='Port number of grpc server (optional).')
@click.option('-v', '--verbose', count=True, help='Verbosity.')
def watch(codes, input, fids, realtype, output, port, verbose):
    if (codes, fids, realtype) == (tuple(), tuple(), None):
        fail_with_usage()

    set_verbosity(verbose)

    codes_len = len(codes)

    if codes_len == 0:
        if input is None:
            fail_with_usage('Either code or input should be given.')
        if not os.path.exists(input):
            fail_with_usage('Given input does not exist.')

        if os.path.isfile(input):
            if input.endswith('.xlsx'):
                import pandas as pd
                df = pd.read_excel(input, dtype='object')
                code_column = '종목코드'
                if code_column in df:
                    codes = df[code_column]
                else:
                    codes = df.iloc[0]
                codes_len = len(codes)
            elif input.endswith('.txt'):
                with open(input) as f:
                    codes = [line.strip() for line in f]
                codes_len = len(codes)
            else:
                fail_with_usage('Unrecognized input type.')

    if realtype is not None:
        from koapy.openapi.RealType import RealType
        fids = list(fids) + RealType.get_fids_by_realtype(realtype)

    if not codes:
        fail_with_usage('No codes to watch. Set --code or --input.')

    if not fids:
        fail_with_usage('Cannot infer fids to watch. Set either --fid or --realtype.')

    import datetime
    import pandas as pd

    from koapy.context.KiwoomOpenApiContext import KiwoomOpenApiContext
    from koapy.openapi.RealType import RealType

    with KiwoomOpenApiContext(port=port, client_check_timeout=client_check_timeout, verbosity=verbose) as context:
        context.EnsureConnected()

        for event in context.WatchRealDataForCodesAsStream(codes, fids, infer_fids=True):
            code = event.listen_response.arguments[0].string_value
            name = event.listen_response.arguments[1].string_value
            fids = event.listen_response.single_data.names
            names = [RealType.Fid.get_name_by_fid(fid, str(fid)) for fid in fids]
            values = event.listen_response.single_data.values
            dic = dict((name, value) for fid, name, value in zip(fids, names, values) if name != fid)
            series = pd.Series(dic)
            click.echo('[%s] [%s]' % (code, name), file=output)
            click.echo('[%s]' % datetime.datetime.now(), file=output)
            click.echo(series.to_markdown(), file=output)

if __name__ == '__main__':
    cli()



import sys

from koapy.pyqt5.KiwoomOpenApiTrayApplication import KiwoomOpenApiTrayApplication

def start_tray_application(args):
    KiwoomOpenApiTrayApplication.main(args)

if __name__ == '__main__':
    start_tray_application(sys.argv)
    


from abc import ABC

from koapy.utils.notimplemented import notimplemented

class KiwoomOpenApiEventHandlerFunctions(ABC):

    @notimplemented
    def OnEventConnect(self, errcode):
        raise NotImplementedError

    @notimplemented
    def OnReceiveMsg(self, scrnno, rqname, trcode, msg):
        raise NotImplementedError

    @notimplemented
    def OnReceiveTrData(self, scrnno, rqname, trcode, recordname, prevnext, datalength, errorcode, message, splmmsg):
        raise NotImplementedError

    @notimplemented
    def OnReceiveRealData(self, code, realtype, realdata):
        raise NotImplementedError

    @notimplemented
    def OnReceiveChejanData(self, gubun, itemcnt, fidlist):
        raise NotImplementedError

    @notimplemented
    def OnReceiveConditionVer(self, ret, msg):
        raise NotImplementedError

    @notimplemented
    def OnReceiveTrCondition(self, scrnno, codelist, condition_name, index, prevnext):
        raise NotImplementedError

    @notimplemented
    def OnReceiveRealCondition(self, code, condition_type, condition_name, condition_index):
        raise NotImplementedError



#!/usr/bin/env python

"""The setup script."""

from setuptools import setup, find_packages

with open('README.rst') as readme_file:
    readme = readme_file.read()

with open('HISTORY.rst') as history_file:
    history = history_file.read()

requirements = [
    'Click>=7.1.2',
    'PyQt5>=5.15.1',
    'grpcio>=1.32.0',
    'grpcio-tools>=1.32.0',
    'protobuf>=3.13.0',
    'pyhocon>=0.3.55',
    'discord.py>=1.4.1',
    'schedule>=0.6.0',
    'tabulate>=0.8.7',
    'tqdm>=4.49.0',
    'wrapt>=1.12.1',
    'numpy>=1.19.2',
    'pandas>=1.1.2',
    'pywin32>=228;sys_platform=="win32"',
    'mslex>=0.1.1;sys_platform=="win32"',
]

setup_requirements = ['pytest-runner', ]

test_requirements = ['pytest>=3', ]

setup(
    author="Yunseong Hwang",
    author_email='kika1492@gmail.com',
    python_requires='>=3.6',
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Natural Language :: Korean',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Operating System :: Microsoft :: Windows :: Windows 10',
        'Environment :: Console',
        'Environment :: X11 Applications :: Qt',
        'Topic :: Office/Business :: Financial',
    ],
    description="Kiwoom Open Api Python",
    entry_points={
        'console_scripts': [
            'koapy=koapy.cli:cli',
        ],
    },
    install_requires=requirements,
    license="MIT license",
    long_description=readme + '\n\n' + history,
    package_data={
        'koapy': [
            'config.conf',
            'openapi/data/fid.xlsx',
            'openapi/data/realtype_by_desc.json',
            'openapi/data/trinfo_by_code.json',
        ]
    },
    include_package_data=True,
    keywords='koapy',
    name='koapy',
    packages=find_packages(include=['koapy', 'koapy.*']),
    setup_requires=setup_requirements,
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/elbakramer/koapy',
    version='0.1.3',
    zip_safe=False,
)



import grpc

from koapy.grpc import KiwoomOpenApiService_pb2_grpc
from koapy.grpc.KiwoomOpenApiServiceClientStubWrapper import KiwoomOpenApiServiceClientStubWrapper
from koapy.config import config

class KiwoomOpenApiServiceClient:

    def __init__(self, host=None, port=None):
        self._host = host or config.get_string('koapy.grpc.host', 'localhost')
        self._port = port or config.get('koapy.grpc.port')

        if self._port is None:
            raise ValueError('Port is None')

        self._target = self._host + ':' + str(self._port)
        self._channel = grpc.insecure_channel(self._target)
        self._stub = KiwoomOpenApiService_pb2_grpc.KiwoomOpenApiServiceStub(self._channel)
        self._stub_wrapper = KiwoomOpenApiServiceClientStubWrapper(self._stub)

    def is_ready(self, timeout=None):
        if timeout is None:
            timeout = config.get_int('koapy.grpc.client.is_ready.timeout', 10)
        try:
            grpc.channel_ready_future(self._channel).result(timeout=timeout)
            return True
        except grpc.FutureTimeoutError:
            return False

    def get_original_stub(self):
        return self._stub

    def get_stub(self):
        return self._stub_wrapper

    def close(self):
        return self._channel.close()

    def __getattr__(self, name):
        return getattr(self._stub_wrapper, name)



import socket

from contextlib import closing
from wrapt import synchronized

@synchronized
def find_free_port_for_host(host=None):
    if host is None:
        host = 'localhost'
    with closing(socket.socket(socket.AF_INET, socket.SOCK_STREAM)) as sock:
        sock.bind((host, 0))
        host, port = sock.getsockname()
        return port

def get_free_localhost_port():
    return find_free_port_for_host('localhost')



import sys
import logging
import argparse
import datetime
import signal
import contextlib

from PyQt5.QtWidgets import QApplication, QSystemTrayIcon, QMenu, QStyle
from PyQt5.QtCore import QTimer, QObject, pyqtSignal

from koapy.grpc.KiwoomOpenApiServiceServer import KiwoomOpenApiServiceServer
from koapy.openapi.KiwoomOpenApiError import KiwoomOpenApiError
from koapy.pyqt5.KiwoomOpenApiQAxWidget import KiwoomOpenApiQAxWidget

from koapy.utils.logging import set_verbosity

class KiwoomOpenApiTrayApplication(QObject):

    _should_restart = pyqtSignal(int)
    _should_restart_exit_code = 1

    def __init__(self, args=()):
        super().__init__()

        self._parser = argparse.ArgumentParser()
        self._parser.add_argument('-p', '--port')
        self._parser.add_argument('--verbose', '-v', action='count', default=0)
        self._parsed_args, remaining_args = self._parser.parse_known_args(args and args[1:])

        self._port = self._parsed_args.port
        self._verbose = self._parsed_args.verbose

        set_verbosity(self._verbose)

        self._app = QApplication(args[:1] + remaining_args)
        self._control = KiwoomOpenApiQAxWidget()
        self._server = KiwoomOpenApiServiceServer(self._control, port=self._port)

        self._should_restart.connect(self._exit)
        self._startRestartNotifier()
        self._startEventLoopProcessor()

        self._tray = QSystemTrayIcon()
        self._tray.activated.connect(self._activate)

        icon = self._app.style().standardIcon(QStyle.SP_TitleBarMenuButton)

        menu = QMenu()
        menu.addSection('Connection')
        connectAction = menu.addAction('Login and Connect')
        connectAction.triggered.connect(self._connect)
        autoLoginAction = menu.addAction('Congifure Auto Login')
        autoLoginAction.triggered.connect(self._configureAutoLogin)
        menu.addSection('Status')
        self._connectionStatusAction = menu.addAction('Status: Disconnected')
        self._connectionStatusAction.setEnabled(False)
        self._serverStatusAction = menu.addAction('Server: Unknown')
        self._serverStatusAction.setEnabled(False)
        menu.addSection('Exit')
        exitAction = menu.addAction('Exit')
        exitAction.triggered.connect(self._exit)

        tooltip = 'Kiwoom OpenAPI Tray Application'

        self._tray.setIcon(icon)
        self._tray.setContextMenu(menu)
        self._tray.setToolTip(tooltip)

        self._tray.show()

        self._control.OnEventConnect.connect(self._OnEventConnect)

    def _checkAndWaitForMaintananceAndThen(self, callback=None, args=None, kwargs=None):
        """
        # 시스템 점검 안내

        안녕하세요. 키움증권 입니다.
        시스템의 안정적인 운영을 위하여
        매일 시스템 점검을 하고 있습니다.
        점검시간은 월~토요일 (05:05 ~ 05:10)
                  일요일    (04:00 ~ 04:30) 까지 입니다.
        따라서 해당 시간대에는 접속단절이 될 수 있습니다.
        참고하시기 바랍니다.
        """
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}

        now = datetime.datetime.now()

        if now.weekday() < 6:
            maintanance_start_time = now.replace(hour=5, minute=5, second=0, microsecond=0)
            maintanance_end_time = now.replace(hour=5, minute=10, second=0, microsecond=0)
        else:
            maintanance_start_time = now.replace(hour=4, minute=0, second=0, microsecond=0)
            maintanance_end_time = now.replace(hour=4, minute=30, second=0, microsecond=0)

        if maintanance_start_time < now < maintanance_end_time:
            target = maintanance_end_time + datetime.timedelta(minutes=5)
            logging.warning('Connection lost due to maintanance, waiting until %s (then will try to reconnect)', target)
            timediff = target - now
            if callback is not None and callable(callback):
                QTimer.singleShot(timediff.total_seconds() * 1000, lambda: callback(*args, **kwargs))

    def _OnEventConnect(self, errcode):
        if errcode == 0:
            logging.debug('Connected to server')
            state = self._control.GetConnectState()
            if state == 1:
                self._connectionStatusAction.setText('Status: Connected')
                server = self._control.GetLoginInfo('GetServerGubun')
                if server == '1':
                    self._serverStatusAction.setText('Server: Simulation')
                else:
                    self._serverStatusAction.setText('Server: Real')
            else:
                raise RuntimeError('Unexpected case')
        elif errcode == KiwoomOpenApiError.OP_ERR_SOCKET_CLOSED:
            logging.error('Socket closed')
            state = self._control.GetConnectState()
            if state == 0:
                self._connectionStatusAction.setText('Status: Disconnected')
            def callback():
                logging.warning('Trying to reconnect')
                self._ensureConnectedAndThen() # 재연결 시도
            self._checkAndWaitForMaintananceAndThen(callback)
        elif errcode == KiwoomOpenApiError.OP_ERR_CONNECT:
            logging.error('Failed to connect')
            state = self._control.GetConnectState()
            if state == 0:
                self._connectionStatusAction.setText('Status: Disconnected')

    def _activate(self, reason):
        if reason == QSystemTrayIcon.DoubleClick:
            self._control.showNormal()
            self._control.activateWindow()

    def _ensureConnectedAndThen(self, callback=None, args=None, kwargs=None):
        if args is None:
            args = ()
        if kwargs is None:
            kwargs = {}
        if self._control.GetConnectState() == 1:
            if callback is not None:
                if not callable(callback):
                    raise TypeError('Function is not callable')
                callback(*args, **kwargs)
        else:
            if callback is not None:
                if not callable(callback):
                    raise TypeError('Function is not callable')
                def callbackAndDisconnect(errcode):
                    self._control.OnEventConnect.disconnect(callbackAndDisconnect)
                    if errcode == 0:
                        callback(*args, **kwargs)
                self._control.OnEventConnect.connect(callbackAndDisconnect)
            self._control.CommConnect()

    def _connect(self):
        self._ensureConnectedAndThen()

    def _showAccountWindow(self):
        self._control.koapy_Functions('ShowAccountWindow', '')

    def _configureAutoLogin(self):
        self._ensureConnectedAndThen(self._showAccountWindow)

    def _onInterrupt(self, signum, _frame):
        self._exit(signum + 100)

    def _exec(self):
        with contextlib.ExitStack() as stack:
            orig_handler = signal.signal(signal.SIGINT, self._onInterrupt)
            stack.callback(signal.signal, signal.SIGINT, orig_handler)
            logging.debug('Starting app')
            logging.debug('Starting server')
            try:
                self._server.start()
            except ValueError as e:
                logging.warning('Error: %s', e)
            logging.debug('Started server')
            return self._app.exec_()

    def _exit(self, return_code=0):
        logging.debug('Exiting app')
        logging.debug('Stopping server')
        self._server.stop(10)
        logging.debug('Waiting for server to stop')
        self._server.wait_for_termination(10)
        logging.debug('Stopped server')
        logging.debug('Quitting app')
        self._app.exit(return_code)
        logging.debug('Quitted app')

    def _nextRestartTime(self):
        now = datetime.datetime.now()
        target = now.replace(hour=6, minute=50, second=0, microsecond=0)
        if now >= target:
            target += datetime.timedelta(days=1)
        return target

    def _startRestartNotifier(self):
        def notify_and_wait_for_next():
            self._should_restart.emit(self._should_restart_exit_code)
            now = datetime.datetime.now()
            next_restart_time = self._nextRestartTime()
            timediff = next_restart_time - now
            QTimer.singleShot((timediff.total_seconds() + 1) * 1000, notify_and_wait_for_next)
        now = datetime.datetime.now()
        next_restart_time = self._nextRestartTime()
        timediff = next_restart_time - now
        QTimer.singleShot((timediff.total_seconds() + 1) * 1000, notify_and_wait_for_next)

    def _startEventLoopProcessor(self):
        interval = 5 * 1000
        def process_and_wait():
            QApplication.processEvents()
            QTimer.singleShot(interval, process_and_wait)
        QTimer.singleShot(interval, process_and_wait)

    def _exitForRestart(self):
        return self._exit(self._should_restart_exit_code)

    def execAndExit(self):
        code = self._exec()
        sys.exit(code)

    def execAndExitWithAutomaticRestart(self):
        should_restart = True
        while should_restart:
            code = self._exec()
            logging.debug('App exitted with return code: %d', code)
            should_restart = code == self._should_restart_exit_code
            if should_restart:
                logging.warning('Exitted app for restart')
                logging.warning('Restarting app')
        sys.exit(code)

    @classmethod
    def main(cls, args):
        app = cls(args)
        app.execAndExitWithAutomaticRestart()



import time
import threading

from functools import wraps

class SimpleRateLimiter:

    """
    [조회횟수 제한 관련 가이드]
      - 1초당 5회 조회를 1번 발생시킨 경우 : 17초대기
      - 1초당 5회 조회를 5연속 발생시킨 경우 : 90초대기
      - 1초당 5회 조회를 10연속 발생시킨 경우 : 3분(180초)대기
    """

    def __init__(self, period, calls):
        self._period = period
        self._calls = calls

        self._lock = threading.RLock()
        self._clock = time.time

        if hasattr(time, 'monotonic'):
            self._clock = time.monotonic

        self._num_calls = 0
        self._last_reset = 0

    def _period_remaining(self):
        elapsed = self._clock() - self._last_reset
        return self._period - elapsed

    def __call__(self, func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            with self._lock:
                period_remaining = self._period_remaining()
                if period_remaining <= 0:
                    self._num_calls = 0
                    self._last_reset = self._clock()
                self._num_calls += 1
                if self._num_calls > self._calls:
                    if period_remaining > 0:
                        time.sleep(period_remaining)
                    self._num_calls = 1
                    self._last_reset = self._clock()
            return func(*args, **kwargs)
        return wrapper
        


import re
import datetime
import requests

def get_krx_closing_dates_as_dict():
    now = datetime.datetime.now()
    def generate_otp():
        headers = {
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate',
            'Host': 'marketdata.krx.co.kr',
            'Referer': 'http://marketdata.krx.co.kr/mdi',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
            'X-Requested-With': 'XMLHttpRequest',
        }
        params = {
            'bld': 'MKD/01/0110/01100305/mkd01100305_01',
            'name': 'form',
            '_': str(int(now.timestamp() * 1000)),
        }
        response = requests.get('http://marketdata.krx.co.kr/contents/COM/GenerateOTP.jspx', headers=headers, params=params)
        code = response.content
        return code
    code = generate_otp()
    headers = {
        'Accept': 'application/json, text/javascript, */*; q=0.01',
        'Accept-Encoding': 'gzip, deflate',
        'Host': 'marketdata.krx.co.kr',
        'Origin': 'http://marketdata.krx.co.kr',
        'Referer': 'http://marketdata.krx.co.kr/mdi',
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
        'X-Requested-With': 'XMLHttpRequest',
    }
    data = {
        'search_bas_yy': str(now.year),
        'gridTp': 'KRX',
        'pagePath': '/contents/MKD/01/0110/01100305/MKD01100305.jsp',
        'code': code,
        'pageFirstCall': 'Y',
    }
    response = requests.post('http://marketdata.krx.co.kr/contents/MKD/99/MKD99000001.jspx', headers=headers, data=data)
    body = response.json()
    return body

def download_krx_closing_dates_as_excel(f=None):
    now = datetime.datetime.now()
    def generate_otp():
        headers = {
            'Accept': '*/*',
            'Accept-Encoding': 'gzip, deflate',
            'Host': 'marketdata.krx.co.kr',
            'Referer': 'http://marketdata.krx.co.kr/mdi',
            'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
            'X-Requested-With': 'XMLHttpRequest',
        }
        params = {
            'name': 'fileDown',
            'filetype': 'xls',
            'url': 'MKD/01/0110/01100305/mkd01100305_01',
            'search_bas_yy': str(now.year),
            'gridTp': 'KRX',
            'pagePath': '/contents/MKD/01/0110/01100305/MKD01100305.jsp',
        }
        response = requests.get('http://marketdata.krx.co.kr/contents/COM/GenerateOTP.jspx', headers=headers, params=params)
        code = response.content
        return code
    code = generate_otp()
    headers = {
        'Accept': 'text/html,application/xhtml+xml,application/xml;q=0.9,image/webp,image/apng,*/*;q=0.8,application/signed-exchange;v=b3;q=0.9',
        'Accept-Encoding': 'gzip, deflate',
        'Accept-Language': 'ko-KR,ko;q=0.9,en-US;q=0.8,en;q=0.7',
        'Cache-Control': 'max-age=0',
        'Host': 'file.krx.co.kr',
        'Origin': 'http://marketdata.krx.co.kr',
        'Referer': 'http://marketdata.krx.co.kr/mdi',
        'Upgrade-Insecure-Requests': '1',
        'User-Agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/84.0.4147.125 Safari/537.36',
    }
    data = {
        'code': code,
    }
    response = requests.post('http://file.krx.co.kr/download.jspx', headers=headers, data=data)
    if f is not None:
        if hasattr(f, 'write'):
            f.write(response.content)
        elif isinstance(f, str):
            filename = f
            with open(filename, 'wb') as f:
                f.write(response.content)
    else:
        filename = re.findall("filename=(.+)", response.headers['Content-Disposition'])[0]
        with open(filename, 'wb') as f:
            f.write(response.content)
    return response
    


import sys

from koapy.openapi.KiwoomOpenApiError import KiwoomOpenApiError

def main():
    codes = sys.argv[1:]
    for code in codes:
        code = int(code)
        print('[%d] %s' % (code, KiwoomOpenApiError.get_error_message_by_code(code)))

if __name__ == '__main__':
    main()



import logging
import threading

class KiwoomOpenApiSignalConnector:

    def __init__(self, name=None):
        self._name = name
        self._lock = threading.RLock()
        self._slots = list()

    def connect(self, slot):
        with self._lock:
            if slot not in self._slots:
                self._slots.append(slot)

    def disconnect(self, slot=None):
        with self._lock:
            if slot is None:
                self._slots.clear()
            elif slot in self._slots:
                self._slots.remove(slot)
            else:
                logging.warning("Tried to disconnect a slot that doesn't exist")

    def get_slots(self):
        with self._lock:
            # make a copy in order to prevent modification during iteration
            # problem
            return list(self._slots)

    def call(self, *args, **kwargs):
        for slot in self.get_slots():
            slot(*args, **kwargs)

    def __call__(self, *args, **kwargs):
        self.call(*args, **kwargs)



import sys

from koapy.openapi.TrInfo import TrInfo

def get_output_names_by_trcode(trcode):
    trinfo = TrInfo.get_trinfo_by_code(trcode)
    single = trinfo.get_single_output_names()
    multi = trinfo.get_multi_output_names()
    return single, multi

def main():
    trcodes = sys.argv[1:]
    for trcode in trcodes:
        print('%s:' % trcode)
        single, multi = get_output_names_by_trcode(trcode)
        print('- single: %r' % single)
        print('- multi: %r' % multi)
        print()

if __name__ == '__main__':
    main()
